//
//  Manager.swift
//  SwiftDMCPExampleSimple
//
//  Created by Malcolm Anderson on 6/4/21.
//

import Foundation
import SwiftDMCP
import SwiftSocket
import CoreMotion
import Combine
import Integreat

class Manager: ObservableObject {
    /// Keeps track of the current tab the user is on (Profiles, Settings, or About)
    @Published var currentIndex: Int = 0
    
    /// The connection to the RMC.
    @Published var connection: DMCPConnection? = nil
    
    /// The IP address of the RMC to connect to.
    @Published var ipAddress: String = "192.168.6.123"
    
    /// Whether or not the device is currently connected to the RMC.
    @Published var connected: Bool = false

    /// Errors thrown by the RMC connection.
    @Published var error: IdentifiableError? = nil
    
    /// The kind of RMC connected to.
    @Published var rmcType: RMCType? = nil
    
    /// The last recorded attitude (rotation) data.
    var attitude: CMAttitude? = nil
    
    /// A value to multiply all three attitude axes by before sending to the RMC.
    @Published var sensitivity: Double = 0.1
    
    /// The current yaw of the device (rotation parallel to the ground), in radians.
    var yaw: Double = 0.0
    
    /// The current pitch of the device (rotation towards and away from the user), in radians.
    var pitch: Double = 0.0
    
    /// The current roll of the device (rotation to the left and right), in radians.
    var roll: Double = 0.0
    
    /// The amount to offset the raw yaw of outputted values by, in degrees.
    @Published var yawOffset: Double = 0.0
    
    /// The amount to offset the raw pitch of outputted values by, in degrees.
    @Published var pitchOffset: Double = 0.0
    
    /// The amount to offset the raw roll of outputted values by, in degrees.
    @Published var rollOffset: Double = 0.0
    
    /// Whether or not to multiply all outputted yaw values by -1.
    @Published var yawNegate: Bool = false
    
    /// Whether or not to multiply all outputted pitch values by -1.
    @Published var pitchNegate: Bool = false
    
    /// Whether or not to multiply all outputted roll values by -1.
    @Published var rollNegate: Bool = false
    
    /// Whether or not to update the value of `yaw`.
    @Published var yawUpdate: Bool = true
    
    /// Whether or not to update the value of `pitch`.
    @Published var pitchUpdate: Bool = true
    
    /// Whether or not to update the value of `roll`.
    @Published var rollUpdate: Bool = true
    
    /// A value to multiply the yaw by before sending to the RMC.
    @Published var yawSensitivity: Double = 1
    
    /// A value to multiply the pitch by before sending to the RMC.
    @Published var pitchSensitivity: Double = 1
    
    /// A value to multiply the roll by before sending to the RMC.
    @Published var rollSensitivity: Double = 1
    
    /// Whether or not to send data to the RMC.
    @Published var sendToRMC: Bool = true
    
    /// Handles retrieval of device motion data.
    var coreMotionManager: CMMotionManager
    
    /// The (desired) number of seconds between device motion updates and sending to the RMC.
    var updateInterval: Double = 0.01
    
    /// Sends data to the RMC.
    var dataTimer: Timer? = nil
    
    /// Retrieves device motion data.
    let updateDataTimer = Timer.publish(every: 0.01, on: .current, in: .common).autoconnect()
    var updateDataCancellable: AnyCancellable? = nil
    
    /**
     Starting address to write to.
     
     The address map is structured in the form
     - inX
     - inY
     - inZ
     - inYaw
     - inPitch
     - inRoll
     - Status (optional)
    
     In the sample at https://forum.deltamotion.com/t/6dof-stewart-platform/478, these addresses
     start at `%MD56.0` and include the Status variable for confirming when the RMC is ready for data.
     
     The current values are set up to align with the "6DOF\_Waldo\_And\_Demo" RMCTools project; here, the addresses start
     at `%MD1024.130`. This project also does not have a Status variable.
     */
    private let address = DMCPAddress(1024, 130)
    
    /// Whether or not to wait for the RMC to signal it's ready for more data before writing.
    private let confirmRMCIsReady = false
    
    /// The list of all motion profiles loaded.
    var motionProfiles: [MotionProfile] = []
    
    /// The currently active motion profile.
    var currentProfile: MotionProfile? = nil
    
    /// The currently calculated motion profile output.
    var output: MotionProfileOutput = MotionProfileOutput()
    
    /// The currently set user parameters for the current motion profile.
    @Published var motionProfileParameters: [Double] = [0,0,0,0,0,0,0,0]
    
    init(connectAutomatically: Bool = false) {
        self.coreMotionManager = CMMotionManager()
        
        // Set up motion profiles
        self.motionProfiles = [
            LiveMotionProfile(self),
            VerticalBounceProfile(),
            MoveInCirclesProfile(),
            SwishProfile(),
            ManualProfile()
        ]
        
        if connectAutomatically {
            self.tryConnect()
        }
    }
    
    /// Attempts to connect to the RMC using the address the user has supplied.
    /// If successful, the user will be moved to the interaction view; if not, an error message
    /// will appear.
    func tryConnect() {
        self.connection = DMCPConnection(address: self.ipAddress)
        print("Try to connect to \(self.ipAddress)")
        
        self.connection!.connect { error in
            print("Tried to connect to \(self.ipAddress) - error was \(String(describing: error))")
            
            // If error is nil, then there was no error - in other words, we're connected
            self.connected = (error == nil)
            
            // If error isn't nil, then display the error to the user
            if error != nil {
                self.error = IdentifiableError(error!)
            } else {
                // Get the RMC type
                self.connection!.getRMCType { error, type in
                    self.rmcType = type
                }
                
                self.beginProcessingMotionUpdates()
            }
        }
        
        self.connection!.onDisconnected = { self.connected = false }
    }
    
    func beginProcessingMotionUpdates() {
        // Steps for getting device attitude at specific interval are from
        // Apple documentation here: https://developer.apple.com/documentation/coremotion/getting_processed_device-motion_data
        // under "Getting Device Motion Data Only When You Need It"
        self.coreMotionManager.showsDeviceMovementDisplay = true
        
        // Update attitude data on the specified interval
        self.coreMotionManager.deviceMotionUpdateInterval = self.updateInterval
        self.coreMotionManager.startDeviceMotionUpdates(using: .xArbitraryCorrectedZVertical)
        
        self.updateDataCancellable = self.updateDataTimer.sink(
            receiveCompletion: { _ in},
            receiveValue: { value in
                // Updating the stored attitude (yaw, pitch, roll) data.
                self.attitude = self.coreMotionManager.deviceMotion?.attitude
                if self.yawUpdate { self.yaw = self.attitude?.yaw ?? 0.0 }
                if self.pitchUpdate { self.pitch = self.attitude?.pitch ?? 0.0 }
                if self.rollUpdate { self.roll = self.attitude?.roll ?? 0.0 }
            }
        )
        
        // In the background, we handle sending data to the RMC, if necessary.
        DispatchQueue.global(qos: .background).async {
            self.dataTimer = Timer(fire: Date(), interval: self.updateInterval, repeats: true) { _ in
                
                // If the user has "paused" communication with the RMC, they don't have a motion
                // profile selected currently, or they're not connected to the RMC, then there's
                // no need to perform any of the upcoming calculations.
                if !self.sendToRMC || self.currentProfile == nil || !self.connected { return }
                
                // Determine what the output should be for the current time.
                let t = Date().timeIntervalSinceReferenceDate
                self.output = self.currentProfile?.evaluate(t, withParameterValues: self.motionProfileParameters) ?? MotionProfileOutput()
                
                // Construct x, y, and z values.
                let x = self.output.x
                let y = self.output.y
                let z = self.output.z
                
                // Construct array of yaw, pitch, and roll values.
                let yawDeg = (self.output.yaw + self.yawOffset)
                    * (self.yawNegate ? -1.0 : 1.0)
                    * self.yawSensitivity
                    * self.sensitivity
                
                let pitchDeg = (self.output.pitch + self.pitchOffset)
                    * (self.pitchNegate ? -1.0 : 1.0)
                    * self.pitchSensitivity
                    * self.sensitivity
                
                let rollDeg = (self.output.roll + self.rollOffset)
                    * (self.rollNegate ? -1.0 : 1.0)
                    * self.rollSensitivity
                    * self.sensitivity
                
                var valueArray: [DMCPData] = [
                    DMCPData.real(Float32(x)),
                    DMCPData.real(Float32(y)),
                    DMCPData.real(Float32(z)),
                    DMCPData.real(Float32(yawDeg)),
                    DMCPData.real(Float32(pitchDeg)),
                    DMCPData.real(Float32(rollDeg))
                ]
                
                if self.confirmRMCIsReady {
                    valueArray.append(DMCPData.int(2))
                    
                    // Check if the RMC is ready for a new set of data.
                    self.connection!.readInt(56, self.address.element + 6) { error, initialControl in
                        if error != nil {
                            switch error! {
                            case SocketError.queryFailed:
                                // I think this is OK, it just means the buffer is full.
                                print("Query failed error")
                                return
                            default:
                                print("\(Date()) - Error reading Control: \(error!)")
                                return
                            }
                        }
    
                        // If the "Ready for Data" bit is set, write the new data.
                        if (initialControl! & 1) != 0 {
                            self.connection!.writeValues(self.address, values: valueArray) { error in
                                if error != nil {
                                    print("\(Date()) - Error writing yaw/pitch/roll: \(error!)")
                                }
                            }
                        }
                    }
                }
                else {
                    self.connection!.writeValues(self.address, values: valueArray) { error in
                        if error != nil {
                            print("\(Date()) - Error writing yaw/pitch/roll: \(error!)")
                        }
                    }
                }
            }
            
            RunLoop.current.add(self.dataTimer!, forMode: RunLoop.Mode.default)
            RunLoop.current.run()
        }
    }
    
    /// Disconnects from the RMC.
    func disconnect() {
        print("Disconnecting from RMC")
        connection?.disconnect()
    }
}

extension Double {
    /// Converts a radian value to degrees.
    func degrees() -> Double {
        return self * 180.0 / .pi
    }
}

/// Simple wrapper class around the `Error` struct that allows us to take advantage
/// of SwiftUI's `Alert` system in a more succinct fashion. With this, we can also package
/// human-readable strings in with the errors for easy display.
class IdentifiableError: Identifiable {
    var id: Error? = nil
    var string: String = ""
    
    init(_ error: Error) {
        self.id = error
        
        var errorString: String
        switch error {
        case SocketError.queryFailed:
            errorString = "The query failed."
        case SocketError.connectionClosed:
            errorString = "The connection was closed by the host."
        case SocketError.connectionTimeout:
            errorString = "The connection timed out."
        case DMCPResponseError.invalidAddress:
            errorString = "The RMC variable address provided was invalid."
        case DMCPResponseError.malformed:
            errorString = "The request was not in the right format."
        case DMCPResponseError.tooLong:
            errorString = "The request was too long."
        default:
            errorString = "An unknown error occurred."
        }
        self.string = errorString + " (\(error))"
    }
}
