//
//  RotationSettingsView.swift
//  SwiftDMCPExample6DOF
//
//  Created by Malcolm Anderson on 6/25/21.
//

import SwiftUI

/**
 Settings page for adjusting individual rotation axes.
 */
struct RotationSettingsView: View {
    @EnvironmentObject var manager: Manager
    var body: some View {
        Form {
            SingleAxisSettingsView(
                header: "Pitch",
                footer: "Refers to tilting the device towards and away from you.",
                color: .red,
                sendToRMC: $manager.pitchUpdate,
                invertValues: $manager.pitchNegate,
                offset: $manager.pitchOffset,
                sensitivity: $manager.pitchSensitivity
            )
            SingleAxisSettingsView(
                header: "Roll",
                footer: "Refers to tilting the device left and right.",
                color: .green,
                sendToRMC: $manager.rollUpdate,
                invertValues: $manager.rollNegate,
                offset: $manager.rollOffset,
                sensitivity: $manager.rollSensitivity
            )
            SingleAxisSettingsView(
                header: "Yaw",
                footer: "Refers to rotating the device parallel to the ground.",
                color: .blue,
                sendToRMC: $manager.yawUpdate,
                invertValues: $manager.yawNegate,
                offset: $manager.yawOffset,
                sensitivity: $manager.yawSensitivity
            )
            
            Section(header: Text("Global"), footer: Text("Refers to all axes")) {
                HStack {
                    Text("Sensitivity \(Int(manager.sensitivity * 100.0))%")
                        .alignmentGuide(.trailing) { d in d[.leading] }
                    Spacer()

                    Slider(
                        value: $manager.sensitivity,
                        in: 0.0...1.0
                    ) {
                        Text("Sensitivity")
                    }
                    
                    Button(action: { manager.sensitivity = 0.5 }) {
                        Text("Reset")
                    }
                    .accentColor(.red)
                }
            }
        }
        .navigationTitle("Axes")
    }
}

struct SingleAxisSettingsView: View {
    @State var header: String = "Header"
    @State var footer: String = "Footer text goes here"
    @State var color: Color
    
    @Binding var sendToRMC: Bool
    @Binding var invertValues: Bool
    @Binding var offset: Double
    @Binding var sensitivity: Double
    
    var body: some View {
        Section(header: Text(header).foregroundColor(color), footer: Text(footer)) {
            
            Toggle(isOn: $sendToRMC) {
                Text("Send to RMC")
            }
            
            Toggle(isOn: $invertValues) {
                Text("Invert Values")
            }
            
            HStack {
                Text("Offset by \(String(format: "%.02f", offset))°")
                
                Slider(
                    value: $offset,
                    in: -180.0...180.0
                ) {
                    Text("Offset")
                }
                
                Button(action: { offset = 0.0 }) {
                    Text("Reset")
                }
                .accentColor(.red)
            }
            
            HStack {
                Text("Sensitivity \(Int(sensitivity * 100.0))%")
                    .alignmentGuide(.trailing) { d in d[.leading] }
                Spacer()

                Slider(
                    value: $sensitivity,
                    in: 0.0...1.0
                ) {
                    Text("Sensitivity")
                }
                
                Button(action: { sensitivity = 0.5 }) {
                    Text("Reset")
                }
                .accentColor(.red)
            }
        }
    }
}

struct RotationSettingsView_Previews: PreviewProvider {
    static var previews: some View {
        RotationSettingsView().environmentObject(Manager())
    }
}
