//
//  SingleProfileView.swift
//  SwiftDMCPExample6DOF
//
//  Created by Malcolm Anderson on 6/28/21.
//

import SwiftUI

/**
 Displays information about the currently selected profile.
 From here, users can see what the values being sent to the RMC
 are, as well as resume/pause sending values, read a short description
 of the motion profile, and adjust any parameters for the profile that may
 exist.
 */
struct SingleProfileView: View {
    @EnvironmentObject var manager: Manager

    @State var profile: MotionProfile
    
    @State var output: MotionProfileOutput = MotionProfileOutput()
    
    var body: some View {
        VStack {
            HStack {
                SingleAxisLabelView(
                    title: "X",
                    color: .red,
                    value: self.$output.x
                )
                SingleAxisLabelView(
                    title: "Y",
                    color: .green,
                    value: self.$output.y
                )
                SingleAxisLabelView(
                    title: "Z",
                    color: .blue,
                    value: self.$output.z
                )
            }
            HStack {
                SingleAxisLabelView(
                    title: "Pitch",
                    suffix: "°",
                    color: .red,
                    value: self.$output.pitch
                )
                SingleAxisLabelView(
                    title: "Roll",
                    suffix: "°",
                    color: .green,
                    value: self.$output.roll
                )
                SingleAxisLabelView(
                    title: "Yaw",
                    suffix: "°",
                    color: .blue,
                    value: self.$output.yaw
                )
            }

            Button(action: { self.manager.sendToRMC.toggle() }) {
                ZStack {
                    Rectangle()
                        .aspectRatio(1.0, contentMode: .fill)
                        .foregroundColor(.clear)
                    Image(systemName: "\(self.manager.sendToRMC ? "pause" : "play").fill")
                        .resizable()
                        .aspectRatio(contentMode: .fit)
                        .foregroundColor(.primary)
                }
                .frame(maxWidth: 60, maxHeight: 60)

            }
            .padding()
            .animation(.default)
            
            VStack(alignment: .leading) {
                HStack {
                    Text("Description")
                        .font(.title).bold()
                    Spacer()
                }
                .padding([.bottom], 5)
                Text(self.profile.description)
            }
            .padding([.bottom], 10)
            
            if self.profile.parameters.count > 0 {
                VStack(alignment: .leading) {
                    HStack {
                        Text("Parameters")
                            .font(.title).bold()
                        Spacer()
                    }
                    .padding([.bottom], 5)
                    ForEach(Array(zip(self.profile.parameters.indices, self.profile.parameters)), id: \.0) { i, profile in
                        VStack(alignment: .leading) {
                            Text("\(profile.name): \(String(format: "%.2f", self.manager.motionProfileParameters[i]))")
                            Slider(
                                value: self.$manager.motionProfileParameters[i],
                                in: profile.minimumValue...profile.maximumValue,
                                minimumValueLabel: Text(String(format: "%.2f", profile.minimumValue)),
                                maximumValueLabel: Text(String(format: "%.2f", profile.maximumValue))) {
                                Text("\(profile.name)")
                            }
                        }
                    }
                }
            }
        }
        .padding()
        .onAppear {
            self.manager.currentProfile = profile
            for (i, parameter) in self.profile.parameters.enumerated() {
                self.manager.motionProfileParameters[i] = parameter.defaultValue
            }
        }
        .onReceive(self.manager.updateDataTimer) { i in
            self.output = self.manager.output
        }
        .onDisappear {
            self.manager.currentProfile = nil
        }
        .navigationTitle(self.profile.name)
    }
}

struct SingleProfileView_Previews: PreviewProvider {
    static var previews: some View {
        SingleProfileView(profile: VerticalBounceProfile()).preferredColorScheme(.dark).environmentObject(Manager())
    }
}
